# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import abc
from typing import BinaryIO, Iterable, Optional

from dyff.schema.platform import Artifact, ArtifactURL, Entities, StorageSignedURL


class StorageBackend(abc.ABC):
    @abc.abstractmethod
    def storage_size(self, path: str) -> int:
        """Get the total size, in bytes, of all objects stored under the ``path``
        prefix."""

    @abc.abstractmethod
    def list_dir(self, path: str, *, recursive: bool = False) -> Iterable[str]:
        """Get the absolute paths of all objects that are immediate "children" of
        ``path``."""

    @abc.abstractmethod
    def download_recursive(self, source: str, destination: str) -> None:
        """Download all objects stored under the path ``source`` to under the local
        directory ``destination``.

        The directory structure is preserved.
        """

    @abc.abstractmethod
    def upload_recursive(self, source: str, destination: str) -> None:
        """Upload all files stored under the local directory ``source`` to under the
        storage path ``destination``.

        The directory structure is preserved.
        """

    @abc.abstractmethod
    def put_object(self, data: bytes | BinaryIO, destination: str) -> None:
        """Upload a single object in memory."""

    @abc.abstractmethod
    def get_object(self, source: str) -> BinaryIO:
        """Download a single object."""

    @abc.abstractmethod
    def signed_url_for_artifact_upload(
        self,
        artifact: Artifact,
        storage_path: str,
        *,
        size_limit_bytes: Optional[int] = None,
    ) -> StorageSignedURL:
        """Create a temporary signed URL that can be used in a PUT request to upload an
        ``Artifact`` directly to storage."""

    @abc.abstractmethod
    def signed_url_for_dataset_upload(
        self,
        dataset_id: str,
        artifact: Artifact,
        *,
        size_limit_bytes: Optional[int] = None,
        storage_path: Optional[str] = None,
    ) -> StorageSignedURL:
        """Create a temporary signed URL that can be used in a PUT request to upload an
        ``Artifact`` directly to storage."""

    @abc.abstractmethod
    def artifact_downlinks(
        self, entity_kind: Entities, entity_id: str
    ) -> Iterable[ArtifactURL]:
        """Create a list of temporary URLs that can be used in GET requests to download
        all of the artifacts associated with an entity."""

    @abc.abstractmethod
    def object_md5hash(self, storage_path: str) -> bytes:
        """Compute the MD5 hash of an object in storage."""

    def artifact_md5hash(self, artifact: Artifact, storage_path: str) -> bytes:
        return self.object_md5hash(f"{storage_path}/{artifact.path}")

    @abc.abstractmethod
    def dataset_artifact_md5hash(
        self, dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
    ) -> bytes:
        """Compute the MD5 hash of a dataset artifact in storage."""



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
from typing import List, Mapping, Union

PathLike = Union[str, bytes, os.PathLike]

YAMLScalar = Union[str, int, float, bool, None]
YAMLType = Union[YAMLScalar, List["YAMLType"], Mapping[str, "YAMLType"]]
YAMLList = List[YAMLType]
YAMLObject = Mapping[str, YAMLType]

__all__ = [
    "PathLike",
    "YAMLList",
    "YAMLObject",
    "YAMLScalar",
    "YAMLType",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
from typing import Collection, Iterable, Optional

import pymongo

from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    Documentation,
    Entities,
    Evaluation,
    Family,
    InferenceService,
    InferenceSession,
    Labeled,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    Resources,
    SafetyCase,
    Tag,
)
from dyff.schema.requests import (
    AuditQueryRequest,
    DatasetQueryRequest,
    DocumentationEditRequest,
    EvaluationQueryRequest,
    InferenceServiceQueryRequest,
    InferenceSessionQueryRequest,
    MeasurementQueryRequest,
    MethodQueryRequest,
    ModelQueryRequest,
    ModuleQueryRequest,
    ReportQueryRequest,
    SafetyCaseQueryRequest,
    TagCreateRequest,
)
from dyff.storage import timestamp
from dyff.storage.backend.base.query import QueryBackend, Whitelist
from dyff.storage.config import config


class MongoDBQueryBackend(QueryBackend):
    def __init__(self):
        connection_string = config.api.query.mongodb.connection_string
        self._client = pymongo.MongoClient(connection_string.get_secret_value())
        self._workflows_db = self._client.get_database(
            config.api.query.mongodb.database
        )

    def _convert_entity_response(self, entity: dict) -> dict:
        def array_to_object(xs: list[dict]) -> dict:
            return {x["key"]: x["value"] for x in xs}

        entity = dict(entity)
        entity["id"] = entity["_id"]
        entity["labels"] = array_to_object(entity.get("labels", []))
        # TODO:
        # entity["annotations"] = array_to_object(entity["annotations"])
        del entity["_id"]
        return entity

    def _get_entity(self, kind: Entities, id: str) -> Optional[dict]:
        collection_name = Resources.for_kind(kind)
        collection = self._workflows_db[collection_name]
        result = collection.find_one({"_id": id})
        if result:
            result = self._convert_entity_response(result)
        return result

    def _query_entities(
        self, kind: Entities, whitelist: Whitelist, query: dict
    ) -> Iterable[dict]:
        print(f"whitelist: {whitelist}")
        mongo_conjunction: list[dict] = []
        if ("*" not in whitelist.accounts) and ("*" not in whitelist.entities):
            # Query constraint requiring the result to be in the whitelist
            mongo_conjunction.append(
                {
                    "$or": [
                        {"account": {"$in": list(whitelist.accounts)}},
                        {"_id": {"$in": list(whitelist.entities)}},
                    ]
                }
            )

        for k, v in query.items():
            if v is None:
                continue
            if k == "id":
                k = "_id"

            if k == "labels":
                labels = json.loads(v)
                labeled = Labeled(labels=labels)  # validate
                for label_key, label_value in labeled.labels.items():
                    mongo_conjunction.append(
                        {k: {"$elemMatch": {"key": label_key, "value": label_value}}}
                    )
            # TODO:
            # elif k == "annotations":
            #     pass
            elif k == "inputsAnyOf":
                # This is a special query for Analysis-related entities. It
                # selects entities that reference any of the listed IDs in
                # their .inputs field.
                entity_list = v.split(",")
                mongo_conjunction.append(
                    {"inputs": {"$elemMatch": {"entity": {"$in": entity_list}}}}
                )
            elif isinstance(v, list):
                mongo_conjunction.append({k: {"$in": list(v)}})
            else:
                mongo_conjunction.append({k: v})

        query = {"$and": mongo_conjunction} if mongo_conjunction else {}
        print(f"query: {query}")
        collection_name = Resources.for_kind(kind)
        collection = self._workflows_db[collection_name]
        results = collection.find(query)

        for result in results:
            yield self._convert_entity_response(result)

    def _rename_query_key(self, query_dict: dict, old_key: str, new_key: str) -> None:
        """If query_dict contains a value for old_key, pop the value and add it back at
        new_key."""
        unset = object()
        value = query_dict.pop(old_key, unset)
        if value is not unset:
            query_dict[new_key] = value

    def get_audit(self, id: str) -> Optional[Audit]:
        """Retrieve an Audit entity.

        Parameters:
          id: The unique key of the Audit.

        Returns:
          The Audit, or None if no Audit with the specified key exists.
        """
        result = self._get_entity(Entities.Audit, id)
        return Audit.parse_obj(result) if result else None

    def query_audits(
        self, whitelist: Whitelist, query: AuditQueryRequest
    ) -> Collection[Audit]:
        """Retrieve all Audit entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Audit entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        results = self._query_entities(Entities.Audit, whitelist, qdict)
        return [Audit.parse_obj(result) for result in results]

    def get_data_source(self, id: str) -> Optional[DataSource]:
        """Retrieve a DataSource entity.

        Parameters:
          id: The unique key of the DataSource.

        Returns:
          The DataSource, or None if no DataSource with the specified key exists.
        """
        result = self._get_entity(Entities.DataSource, id)
        return DataSource.parse_obj(result) if result else None

    def query_data_sources(
        self, whitelist: Whitelist, **query
    ) -> Collection[DataSource]:
        """Retrieve all DataSource entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the DataSource entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        results = self._query_entities(Entities.DataSource, whitelist, query)
        return [DataSource.parse_obj(result) for result in results]

    def get_dataset(self, id: str) -> Optional[Dataset]:
        """Retrieve a Dataset entity.

        Parameters:
          id: The unique key of the Dataset.

        Returns:
          The Dataset, or None if no Dataset with the specified key exists.
        """
        result = self._get_entity(Entities.Dataset, id)
        return Dataset.parse_obj(result) if result else None

    def query_datasets(
        self, whitelist: Whitelist, query: DatasetQueryRequest
    ) -> Collection[Dataset]:
        """Retrieve all Dataset entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Dataset entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        results = self._query_entities(Entities.Dataset, whitelist, qdict)
        return [Dataset.parse_obj(result) for result in results]

    def edit_documentation(
        self, id: str, edit: DocumentationEditRequest
    ) -> Optional[Documentation]:
        """Edit the Documentation entity associated with a resource.

        Parameters:
          id: ID of the documented resource.
          edit: Edit request containing changes to make to the documentation.

        Returns:
          The new Documentation, or None if no resource exists with the given ID.
        """
        collection = self._workflows_db["documentation"]
        # Users can explicitly set a field to None, but we don't want to
        # overwrite with None implicitly
        edit_dict = edit.dict(exclude_unset=True)
        result = collection.find_one_and_update(
            {"_id": id},
            {"$set": edit_dict},
            upsert=True,
            return_document=pymongo.ReturnDocument.AFTER,
        )
        if result is None:
            return None
        del result["_id"]
        return Documentation.parse_obj(result)

    def get_documentation(self, id: str) -> Optional[Documentation]:
        """Retrieve the Documentation entity associated with a resource.

        Parameters:
          id: ID of the documented resource.

        Returns:
          The Documentation, or None if no resource exists with the given ID.
        """
        collection = self._workflows_db["documentation"]
        result = collection.find_one({"_id": id})
        if result is None:
            return None
        del result["_id"]
        return Documentation.parse_obj(result)

    def get_evaluation(self, id: str) -> Optional[Evaluation]:
        """Retrieve an Evaluation entity.

        Parameters:
          id: The unique key of the Evaluation.

        Returns:
          The Evaluation, or None if no Evaluation with the specified key exists.
        """
        result = self._get_entity(Entities.Evaluation, id)
        return Evaluation.parse_obj(result) if result else None

    def query_evaluations(
        self, whitelist: Whitelist, query: EvaluationQueryRequest
    ) -> Collection[Evaluation]:
        """Retrieve all Evaluation entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Evaluation entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(
            qdict, "inferenceService", "inferenceSession.inferenceService.id"
        )
        self._rename_query_key(
            qdict, "inferenceServiceName", "inferenceSession.inferenceService.name"
        )
        self._rename_query_key(
            qdict, "model", "inferenceSession.inferenceService.model.id"
        )
        self._rename_query_key(
            qdict, "modelName", "inferenceSession.inferenceService.model.name"
        )
        results = self._query_entities(Entities.Evaluation, whitelist, qdict)
        return [Evaluation.parse_obj(result) for result in results]

    def get_family(self, id: str) -> Optional[Family]:
        """Retrieve a Family entity.

        Parameters:
          id: The unique ID of the Family.

        Returns:
          The Family, or None if no Family with the specified ID exists.
        """
        result = self._get_entity(Entities.Family, id)
        return Family.parse_obj(result) if result else None

    def get_inference_service(self, id: str) -> Optional[InferenceService]:
        """Retrieve an InferenceService entity.

        Parameters:
          id: The unique key of the InferenceService.

        Returns:
          The InferenceService, or None if no InferenceService with the specified key exists.
        """
        result = self._get_entity(Entities.InferenceService, id)
        return InferenceService.parse_obj(result) if result else None

    def query_inference_services(
        self, whitelist: Whitelist, query: InferenceServiceQueryRequest
    ) -> Collection[InferenceService]:
        """Retrieve all InferenceService entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceService entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(qdict, "model", "model.id")
        self._rename_query_key(qdict, "modelName", "model.name")
        results = self._query_entities(Entities.InferenceService, whitelist, qdict)
        return [InferenceService.parse_obj(result) for result in results]

    def get_inference_session(self, id: str) -> Optional[InferenceSession]:
        """Retrieve an InferenceSession entity.

        Parameters:
          id: The unique key of the InferenceSession.

        Returns:
          The InferenceSession, or None if no InferenceSession with the specified key exists.
        """
        result = self._get_entity(Entities.InferenceSession, id)
        return InferenceSession.parse_obj(result) if result else None

    def query_inference_sessions(
        self, whitelist: Whitelist, query: InferenceSessionQueryRequest
    ) -> Collection[InferenceSession]:
        """Retrieve all InferenceSession entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceSession entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(qdict, "inferenceService", "inferenceService.id")
        self._rename_query_key(qdict, "inferenceServiceName", "inferenceService.name")
        self._rename_query_key(qdict, "model", "inferenceService.model.id")
        self._rename_query_key(qdict, "modelName", "inferenceService.model.name")
        results = self._query_entities(Entities.InferenceSession, whitelist, qdict)
        return [InferenceSession.parse_obj(result) for result in results]

    def get_measurement(self, id: str) -> Optional[Measurement]:
        """Retrieve a Measurement entity.

        Parameters:
          id: The unique key of the Measurement.

        Returns:
          The Measurement, or None if no Measurement with the specified key exists.
        """
        result = self._get_entity(Entities.Measurement, id)
        return Measurement.parse_obj(result) if result else None

    def query_measurements(
        self, whitelist: Whitelist, query: MeasurementQueryRequest
    ) -> Collection[Measurement]:
        """Retrieve all Measurement entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Measurement entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(qdict, "method", "method.id")
        self._rename_query_key(qdict, "methodName", "method.name")
        self._rename_query_key(qdict, "dataset", "scope.dataset")
        self._rename_query_key(qdict, "evaluation", "scope.evaluation")
        self._rename_query_key(qdict, "inferenceService", "scope.inferenceService")
        self._rename_query_key(qdict, "model", "scope.model")
        results = self._query_entities(Entities.Measurement, whitelist, qdict)
        return [Measurement.parse_obj(result) for result in results]

    def get_method(self, id: str) -> Optional[Method]:
        """Retrieve a Method entity.

        Parameters:
          id: The unique key of the Method.

        Returns:
          The Method, or None if no Method with the specified key exists.
        """
        result = self._get_entity(Entities.Method, id)
        return Method.parse_obj(result) if result else None

    def query_methods(
        self, whitelist: Whitelist, query: MethodQueryRequest
    ) -> Collection[Method]:
        """Retrieve all Method entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Method entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(qdict, "outputKind", "output.kind")
        results = self._query_entities(Entities.Method, whitelist, qdict)
        return [Method.parse_obj(result) for result in results]

    def get_model(self, id: str) -> Optional[Model]:
        """Retrieve a Model entity.

        Parameters:
          id: The unique key of the Model.

        Returns:
          The Model, or None if no Model with the specified key exists.
        """
        result = self._get_entity(Entities.Model, id)
        return Model.parse_obj(result) if result else None

    def query_models(
        self, whitelist: Whitelist, query: ModelQueryRequest
    ) -> Collection[Model]:
        """Retrieve all Model entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Model entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        results = self._query_entities(Entities.Model, whitelist, qdict)
        return [Model.parse_obj(result) for result in results]

    def get_module(self, id: str) -> Optional[Module]:
        """Retrieve a Module entity.

        Parameters:
          id: The unique key of the Module.

        Returns:
          The Module, or None if no Module with the specified key exists.
        """
        result = self._get_entity(Entities.Module, id)
        return Module.parse_obj(result) if result else None

    def query_modules(
        self, whitelist: Whitelist, query: ModuleQueryRequest
    ) -> Collection[Module]:
        """Retrieve all Module entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Module entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        results = self._query_entities(Entities.Module, whitelist, qdict)
        return [Module.parse_obj(result) for result in results]

    def get_report(self, id: str) -> Optional[Report]:
        """Retrieve a Report entity.

        Parameters:
          id: The unique key of the Report.

        Returns:
          The Report, or None if no Report with the specified key exists.
        """
        result = self._get_entity(Entities.Report, id)
        return Report.parse_obj(result) if result else None

    def query_reports(
        self, whitelist: Whitelist, query: ReportQueryRequest
    ) -> Collection[Report]:
        """Retrieve all Report entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Report entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        results = self._query_entities(Entities.Report, whitelist, qdict)
        return [Report.parse_obj(result) for result in results]

    def get_safetycase(self, id: str) -> Optional[SafetyCase]:
        """Retrieve a SafetyCase entity.

        Parameters:
          id: The unique key of the SafetyCase.

        Returns:
          The SafetyCase, or None if no SafetyCase with the specified key exists.
        """
        result = self._get_entity(Entities.SafetyCase, id)
        return SafetyCase.parse_obj(result) if result else None

    def query_safetycases(
        self, whitelist: Whitelist, query: SafetyCaseQueryRequest
    ) -> Collection[SafetyCase]:
        """Retrieve all SafetyCase entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the SafetyCase entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """
        qdict = query.dict()
        self._rename_query_key(qdict, "method", "method.id")
        self._rename_query_key(qdict, "methodName", "method.name")
        self._rename_query_key(qdict, "dataset", "scope.dataset")
        self._rename_query_key(qdict, "evaluation", "scope.evaluation")
        self._rename_query_key(qdict, "inferenceService", "scope.inferenceService")
        self._rename_query_key(qdict, "model", "scope.model")
        results = self._query_entities(Entities.SafetyCase, whitelist, qdict)
        return [SafetyCase.parse_obj(result) for result in results]

    def get_tag(self, id: str) -> Optional[Tag]:
        """Retrieve a Tag entity.

        Parameters:
          id: The unique ID of the Tag.

        Returns:
          The Tag, or None if no Tag with the specified ID exists.
        """
        result = self._get_entity(Entities.Tag, id)
        return Tag.parse_obj(result) if result else None

    def create_tag(self, family: str, tag_request: TagCreateRequest) -> Optional[Tag]:
        """Create a new Tag in an existing Family. Has no effect and returns None if a
        tag with the same name already exists.

        Parameters:
          family: The ID of the Family to add the tag to.
          request: TagCreateRequest

        Returns:
          The created Tag, or None if no Tag was created. This could be because
          the Family doesn't exist, or because a tag with the same name already
          exists in the family.
        """
        tag_dict = tag_request.dict()
        tag_dict["created"] = timestamp.now()
        new_tag = Tag.parse_obj(tag_dict)
        families_collection = self._workflows_db[Resources.Family]
        # Atomically check if the tag exists and add it only if it does not exist
        result = families_collection.update_one(
            {
                # Specified Family
                "_id": family,
                # Tag does not exist in .tags list
                "tags": {"$not": {"$elemMatch": {"tag": new_tag.tag}}},
            },
            {"$push": {"tags": new_tag.dict()}},
        )
        if result.modified_count > 0:
            return new_tag
        else:
            return None



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

# mypy: disable-error-code="import-untyped"
import functools
from typing import BinaryIO, Iterable, Optional

from dyff.schema.platform import Artifact, ArtifactURL, Entities, StorageSignedURL
from dyff.storage import dynamic_import, paths
from dyff.storage.backend.base.storage import StorageBackend
from dyff.storage.config import config


@functools.lru_cache()
def _get_backend() -> StorageBackend:
    return dynamic_import.instantiate(config.storage.backend)


def storage_size(path: str) -> int:
    return _get_backend().storage_size(path)


def list_dir(path: str, *, recursive: bool = False) -> Iterable[str]:
    return _get_backend().list_dir(path, recursive=recursive)


def download_recursive(source: str, destination: str) -> None:
    return _get_backend().download_recursive(source, destination)


def upload_recursive(source: str, destination: str) -> None:
    return _get_backend().upload_recursive(source, destination)


def put_object(data: bytes | BinaryIO, destination: str) -> None:
    return _get_backend().put_object(data, destination)


def get_object(source: str) -> BinaryIO:
    return _get_backend().get_object(source)


def signed_url_for_artifact_upload(
    artifact: Artifact,
    storage_path: str,
    *,
    size_limit_bytes: Optional[int] = None,
) -> StorageSignedURL:
    return _get_backend().signed_url_for_artifact_upload(
        artifact,
        storage_path,
        size_limit_bytes=size_limit_bytes,
    )


def signed_url_for_dataset_upload(
    dataset_id: str,
    artifact: Artifact,
    *,
    size_limit_bytes: Optional[int] = None,
    storage_path: Optional[str] = None,
) -> StorageSignedURL:
    return _get_backend().signed_url_for_dataset_upload(
        dataset_id,
        artifact,
        size_limit_bytes=size_limit_bytes,
        storage_path=storage_path,
    )


def artifact_downlinks(entity_kind: Entities, entity_id: str) -> Iterable[ArtifactURL]:
    return _get_backend().artifact_downlinks(entity_kind, entity_id)


def object_md5hash(storage_path: str) -> bytes:
    return _get_backend().object_md5hash(storage_path)


def artifact_md5hash(artifact: Artifact, storage_path: str) -> bytes:
    return _get_backend().artifact_md5hash(artifact, storage_path)


def dataset_artifact_md5hash(
    dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
) -> bytes:
    return _get_backend().dataset_artifact_md5hash(
        dataset_id, artifact_path, storage_path=storage_path
    )


__all__ = [
    "paths",
    "artifact_downlinks",
    "artifact_md5hash",
    "dataset_artifact_md5hash",
    "download_recursive",
    "get_object",
    "list_dir",
    "object_md5hash",
    "put_object",
    "signed_url_for_artifact_upload",
    "signed_url_for_dataset_upload",
    "storage_size",
    "upload_recursive",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Optional, TypeVar

from dyff.schema import ids
from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    DyffEntity,
    EntityStatus,
    Evaluation,
    InferenceService,
    InferenceSession,
    Labeled,
    Model,
    Report,
)
from dyff.storage import timestamp

_EntityT = TypeVar("_EntityT", bound=DyffEntity)


class MockCommandBackend(abc.ABC):
    def _copy_and_add_system_fields(self, entity: _EntityT) -> _EntityT:
        entity = entity.copy()
        entity.creationTime = timestamp.now()
        entity.status = EntityStatus.created
        entity.id = ids.generate_entity_id()
        return entity

    def create_audit(self, spec: Audit) -> Audit:
        print(f"create_audit\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_data_source(self, spec: DataSource) -> DataSource:
        print(f"create_data_source\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_dataset(self, spec: Dataset) -> Dataset:
        print(f"create_dataset\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_evaluation(self, spec: Evaluation) -> Evaluation:
        print(f"create_evaluation\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_inference_service(self, spec: InferenceService) -> InferenceService:
        print(f"create_inference_service\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_inference_session(self, spec: InferenceSession) -> InferenceSession:
        print(f"create_inference_session\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_model(self, spec: Model) -> Model:
        print(f"create_model\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def create_report(self, spec: Report) -> Report:
        print(f"create_report\nspec: {spec.json(indent=2)}", flush=True)
        return self._copy_and_add_system_fields(spec)

    def update_status(
        self, id: str, *, status: str, reason: Optional[str] = None
    ) -> None:
        print(f"update_status: {id} {status} {reason}", flush=True)

    def update_labels(self, id: str, labels: Labeled) -> None:
        print(f"update_labels: {id}\n{labels.json(indent=2)}", flush=True)



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import base64
import datetime
from pathlib import Path
from typing import Optional, Tuple

import google.auth.compute_engine
import google.auth.transport.requests
import google.cloud.storage as gcs
from google.cloud.storage import transfer_manager

from dyff.schema.dataset import max_artifact_size_bytes
from dyff.schema.platform import Artifact, StorageSignedURL
from dyff.storage import paths
from dyff.storage.backend.base.storage import StorageBackend


def _bucket_name_from_path(path: str) -> str:
    return _split_bucket_path(path)[0]


def _drop_protocol(path: str) -> str:
    protocol = "gs://"
    if path.startswith(protocol):
        path = path[len(protocol) :]
    return path


def _split_bucket_path(path: str) -> Tuple[str, str]:
    path = _drop_protocol(path)
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])


class GCloudStorageBackend(StorageBackend):
    def storage_size(self, path: str) -> int:
        bucket, prefix = _split_bucket_path(path)
        client = gcs.Client()
        bucket_obj = client.get_bucket(bucket)
        blobs = bucket_obj.list_blobs(prefix=prefix)
        return sum(b.size for b in blobs)

    def list_dir(self, path: str) -> list[str]:
        if not path.startswith("gs://"):
            raise ValueError("path must be a GCS object")
        if not path.endswith("/"):
            path += "/"
        client = gcs.Client()
        remote_path = path[len("gs://") :]
        bucket_name, *rest = remote_path.split("/")
        prefix = "/".join(rest)
        bucket = client.get_bucket(bucket_name)
        # Get the objects under the 'source' path
        blobs = bucket.list_blobs(prefix=prefix)
        return [f"gs://{blob.bucket.name}/{blob.name}" for blob in blobs]

    def download_recursive(self, source: str, destination: str) -> None:
        if not source.startswith("gs://"):
            raise ValueError("source must be a GCS object")
        if not source.endswith("/"):
            source += "/"
        client = gcs.Client()
        remote_path = source[len("gs://") :]
        parts = remote_path.split("/")
        bucket_name = parts[0]
        bucket = client.bucket(bucket_name)
        source_blob_name = "/".join(parts[1:])
        bucket = client.get_bucket(bucket_name)
        # Get the objects under the 'source' path
        blobs = bucket.list_blobs(prefix=source_blob_name)
        for blob in blobs:
            if blob.name.endswith("/"):
                # Not a file
                continue
            remote_file = Path(blob.name).relative_to(source_blob_name)
            remote_directory = Path(blob.name).parent.relative_to(source_blob_name)
            local_path = Path(destination) / remote_directory
            local_path.mkdir(parents=True, exist_ok=True)
            blob.download_to_filename(Path(destination) / remote_file)

    def upload_recursive(self, source: str, destination: str) -> None:
        if not destination.startswith("gs://"):
            raise ValueError("destination must be a GCS object")
        if not destination.endswith("/"):
            destination += "/"

        source_path = Path(source)
        if not source_path.is_dir():
            raise ValueError("source must be a local directory")

        storage_client = gcs.Client()
        bucket_name, bucket_path = _split_bucket_path(destination)
        destination_bucket = storage_client.bucket(bucket_name)

        source_paths = [path for path in source_path.rglob("*") if path.is_file()]
        relative_paths = [path.relative_to(source) for path in source_paths]
        string_paths = [str(path) for path in relative_paths]

        if bucket_path != "":
            assert bucket_path[-1] == "/"  # See above
            prefix = bucket_path
        else:
            prefix = ""
        results = transfer_manager.upload_many_from_filenames(
            destination_bucket,
            string_paths,
            source_directory=source,
            blob_name_prefix=prefix,
        )

        exceptions: list[Exception] = []
        exception_names: list[str] = []
        for name, result in zip(string_paths, results):
            if isinstance(result, Exception):
                exceptions.append(result)
                exception_names.append(name)
        if exceptions:
            raise RuntimeError(f"upload failed: {exception_names}") from exceptions[0]

    def signed_url_for_dataset_upload(
        self,
        dataset_id: str,
        artifact: Artifact,
        *,
        size_limit_bytes: Optional[int] = None,
        storage_path: Optional[str] = None,
    ) -> StorageSignedURL:
        if size_limit_bytes is None:
            size_limit_bytes = max_artifact_size_bytes()
        if artifact.digest.md5 is None:
            raise ValueError("requires artifact.digest.md5")
        storage_path = storage_path or paths.dataset_root(dataset_id)
        bucket_name = _bucket_name_from_path(storage_path)
        client = gcs.Client()
        blob = client.bucket(bucket_name).blob(f"{dataset_id}/{artifact.path}")

        auth_request = google.auth.transport.requests.Request()
        signing_credentials = google.auth.compute_engine.IDTokenCredentials(
            auth_request,
            "api-server.dyff.io",
            service_account_email="api-server@dyff-354017.iam.gserviceaccount.com",
        )

        # Google custom header limiting the size of the artifact
        headers = {
            "x-goog-content-length-range": f"0,{size_limit_bytes}",
        }
        url = blob.generate_signed_url(
            version="v4",
            expiration=datetime.timedelta(hours=1),
            method="PUT",
            content_md5=artifact.digest.md5,
            headers=headers.copy(),  # The function mutates this argument
            credentials=signing_credentials,
        )
        return StorageSignedURL(url=url, method="PUT", headers=headers)

    def dataset_artifact_md5hash(
        self, dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
    ) -> bytes:
        storage_path = storage_path or paths.dataset_root(dataset_id)
        bucket_name = _bucket_name_from_path(storage_path)
        client = gcs.Client()
        blob = client.bucket(bucket_name).get_blob(f"{dataset_id}/{artifact_path}")
        if blob is None:
            raise ValueError(
                f"no artifact {artifact_path} stored for dataset {dataset_id}"
            )
        return base64.b64decode(blob.md5_hash)


__all__ = ["GCloudStorageBackend"]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import json
from typing import Optional, TypeVar

import pydantic
from confluent_kafka import Producer

from dyff.schema import ids
from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    DyffEntity,
    DyffEntityType,
    EntityStatus,
    EntityStatusReason,
    Evaluation,
    InferenceService,
    InferenceSession,
    Labeled,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    SafetyCase,
)
from dyff.storage import timestamp
from dyff.storage.config import config
from dyff.storage.typing import YAMLObject

from ..base.command import CommandBackend


def serialize_id(id: str) -> bytes:
    return id.encode("utf-8")


def serialize_value(value: YAMLObject) -> bytes:
    return json.dumps(value).encode("utf-8")


def serialize_entity(entity: DyffEntity) -> bytes:
    return entity.json().encode("utf-8")


def deserialize_id(id: bytes) -> str:
    return id.decode("utf-8")


def deserialize_value(value: bytes) -> YAMLObject:
    return json.loads(value.decode("utf-8"))


def deserialize_entity(entity: bytes) -> DyffEntityType:
    s = entity.decode("utf-8")
    # DyffEntityType is a pydantic "discriminated union", so pydantic infers
    # the type from the '.kind' field
    return pydantic.parse_raw_as(DyffEntityType, s)  # type: ignore


EntityT = TypeVar("EntityT", bound=DyffEntity)


class KafkaCommandBackend(CommandBackend):
    def __init__(self):
        self.events_topic = config.kafka.topics.workflows_events
        producer_config = config.kafka.config.get_producer_config()
        # FIXME: This should be the global default. It's supposed to be the
        # default in Kafka > 3.0, but the docs of librdkafka suggest that
        # its default is still False.
        # See: https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
        producer_config["enable.idempotence"] = True
        # FIXME: See comment in alignmentlabs.dyff.web.server
        # Can't get real logging to work when running in uvicorn
        print(
            f"Creating KafkaCommandBackend with config:\n{json.dumps(producer_config, indent=2)}",
            flush=True,
        )
        # logging.info(f"Creating KafkaCommandBackend with config:\n{json.dumps(producer_config, indent=2)}")
        self._kafka_producer = Producer(producer_config)

    def _copy_and_add_system_fields(self, entity: EntityT) -> EntityT:
        entity = entity.copy()
        entity.creationTime = timestamp.now()
        entity.status = EntityStatus.created
        entity.id = ids.generate_entity_id()
        return entity

    def _produce_create_event(self, entity: EntityT) -> EntityT:
        entity = self._copy_and_add_system_fields(entity)
        assert entity.id is not None
        self._kafka_producer.produce(
            topic=self.events_topic,
            value=serialize_entity(entity),
            key=serialize_id(entity.id),
        )
        return entity

    def _produce_terminate_event(self, id: str) -> None:
        update = {
            "status": EntityStatus.terminated,
            "reason": EntityStatusReason.terminate_command,
        }
        self._kafka_producer.produce(
            topic=self.events_topic,
            value=serialize_value(update),
            key=serialize_id(id),
        )

    def update_status(
        self, id: str, *, status: str, reason: Optional[str] = None
    ) -> None:
        """Update the status of an entity.

        Parameters:
          id: The entity .id
          status: New .status value
          reason: New .reason value
        """
        update = {
            "status": status,
            "reason": reason,
        }
        self._kafka_producer.produce(
            topic=self.events_topic,
            value=serialize_value(update),
            key=serialize_id(id),
        )

    def update_labels(self, id: str, labels: Labeled) -> None:
        """Updated the labels of a labeled entity.

        :param id: The ID of the entity to update.
        :type id: str
        :param labels: The labels to update.
        :type labels: Labeled
        """
        update = labels.dict()
        self._kafka_producer.produce(
            topic=self.events_topic, value=serialize_value(update), key=serialize_id(id)
        )

    def create_audit(self, spec: Audit) -> Audit:
        """Create a new Audit entity in the system.

        Parameters:
          spec: Specification of the Audit. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_data_source(self, spec: DataSource) -> DataSource:
        """Create a new DataSource entity in the system.

        Parameters:
          spec: Specification of the DataSource. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_dataset(self, spec: Dataset) -> Dataset:
        """Create a new Dataset entity in the system.

        Parameters:
          spec: Specification of the Dataset. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_evaluation(self, spec: Evaluation) -> Evaluation:
        """Create a new Evaluation entity in the system.

        Parameters:
          spec: Specification of the Evaluation. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_inference_service(self, spec: InferenceService) -> InferenceService:
        """Create a new InferenceService entity in the system.

        Parameters:
          spec: Specification of the InferenceService. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_inference_session(self, spec: InferenceSession) -> InferenceSession:
        """Create a new InferenceSession entity in the system.

        Parameters:
          spec: Specification of the InferenceSession. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_measurement(self, spec: Measurement) -> Measurement:
        """Create a new Measurement entity in the system.

        Parameters:
          spec: Specification of the Measurement. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_method(self, spec: Method) -> Method:
        """Create a new Method entity in the system.

        Parameters:
          spec: Specification of the Method. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_model(self, spec: Model) -> Model:
        """Create a new Model entity in the system.

        Parameters:
          spec: Specification of the Model. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_module(self, spec: Module) -> Module:
        """Create a new Module entity in the system.

        Parameters:
          spec: Specification of the Module. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_report(self, spec: Report) -> Report:
        """Create a new Report entity in the system.

        Parameters:
          spec: Specification of the Report. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)

    def create_safetycase(self, spec: SafetyCase) -> SafetyCase:
        """Create a new SafetyCase entity in the system.

        Parameters:
          spec: Specification of the SafetyCase. The system fields of the spec
            such as ``.id`` must be **unset**.

        Returns:
          A copy of ``spec`` with all system fields set.
        """
        return self._produce_create_event(spec)



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Optional

import pymongo
import pymongo.read_concern
import pymongo.read_preferences
import pymongo.write_concern
from pymongo.client_session import ClientSession

from dyff.schema import ids
from dyff.schema.platform import APIKey
from dyff.storage import timestamp
from dyff.storage.backend.base.auth import Account, AuthBackend, Identity
from dyff.storage.config import config
from dyff.storage.exceptions import EntityExistsError


class MongoDBAuthBackend(AuthBackend):
    def __init__(self):
        connection_string = config.api.auth.mongodb.connection_string
        self._client = pymongo.MongoClient(connection_string.get_secret_value())

        # Interact with the database in a way that gives strong consistency
        # (These are currently the default settings; I prefer to be explicit)
        self._accounts_db = self._client.get_database(
            config.api.auth.mongodb.database,
            read_concern=pymongo.read_concern.ReadConcern("majority"),
            read_preference=pymongo.ReadPreference.PRIMARY,
            write_concern=pymongo.write_concern.WriteConcern("majority", wtimeout=5000),
        )

    def _get_account(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        identity: Optional[Identity] = None,
        session: Optional[ClientSession] = None,
    ) -> Optional[Account]:
        filter = {}
        if id is not None:
            filter["_id"] = id
        if name is not None:
            filter["name"] = name
        if identity is not None:
            identity_dict = identity.dict()
            for identity_key, identity_value in identity_dict.items():
                if identity_value is not None:
                    filter[f"identity.{identity_key}"] = identity_value
        if len(filter) == 0:
            raise ValueError("must specify at least one of {id, name, identity}")
        result = self._accounts_db.accounts.find_one(filter, session=session)
        if result:
            result = dict(result)
            result["id"] = result["_id"]
            del result["_id"]
            return Account.parse_obj(result)
        return None

    def _insert_account(
        self, account: Account, *, session: Optional[ClientSession] = None
    ) -> None:
        d = account.dict()
        d["_id"] = d["id"]
        del d["id"]
        self._accounts_db.accounts.insert_one(d, session=session)

    def create_account(self, name: str, identity: Optional[Identity] = None) -> Account:
        """Create a new Account.

        Parameters:
        name: A unique human-readable name
        identity: Identity from an external identity provider.

        Returns:
        Entity representing the Account.
        """
        identity = identity or Identity()
        duplicate = self._get_account(name=name, identity=identity)
        if duplicate is not None:
            raise EntityExistsError("duplicate account name or identity")
        id = ids.generate_entity_id()
        account = Account(
            id=id, name=name, identity=identity, creationTime=timestamp.now()
        )
        self._insert_account(account)
        return account

    def delete_account(self, account_id: str):
        """Delete an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        raise NotImplementedError()

    def get_account(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        identity: Optional[Identity] = None,
    ) -> Optional[Account]:
        """Get an Account by ID, identity, or name.

        Parameters:
        id: The unique identifier of the Account.
        name: The unique name of the Account.
        identity: Unique identifier of the Account according to a third-party
          identity provider. It is an error to set more than one identity.
        """
        return self._get_account(id=id, name=name, identity=identity)

    def add_api_key(self, account_id: str, api_key: APIKey) -> None:
        """Add a new APIKey to an Account.

        Raises ``ValueError`` if an APIKey with the same ID already exists.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The new API key.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        if any(account_key.id == api_key.id for account_key in account.apiKeys):
            raise ValueError(f"APIKey with .id {api_key.id} already exists")
        # Don't have to translate APIKey.id because APIKey isn't a
        # top-level entity
        self._accounts_db.accounts.update_one(
            {"_id": account_id}, {"$push": {"apiKeys": api_key.dict()}}
        )

    def revoke_api_key(self, account_id: str, api_key_id: str) -> None:
        """Revoke an APIKey associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The unique identifier of the APIKey.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        filtered = [api_key for api_key in account.apiKeys if api_key.id != api_key_id]
        if len(filtered) == len(account.apiKeys):
            raise ValueError(f"no APIKey with .id {api_key_id} in account")
        # Don't have to translate APIKey.id because APIKey isn't a
        # top-level entity
        self._accounts_db.accounts.update_one(
            {"_id": account_id},
            {"$set": {"apiKeys": [key.dict() for key in filtered]}},
        )

    def revoke_all_api_keys(self, account_id: str) -> None:
        """Revoke all API keys for the given Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        result = self._accounts_db["accounts"].update_one(
            {"_id": account_id}, {"$set": {"apiKeys": []}}
        )
        if result.matched_count != 1:
            raise ValueError(f"no Account with .id {account_id}")

    def get_api_key(self, account_id: str, api_key_id: str) -> Optional[APIKey]:
        """Get an APIKey associated with an Account by ID.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key_id: The unique identifier of the APIKey.
        """
        api_keys = self.get_all_api_keys(account_id)
        for api_key in api_keys:
            if api_key.id == api_key_id:
                return api_key
        return None

    def get_all_api_keys(self, account_id: str) -> list[APIKey]:
        """Get the API keys associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        account = Account.parse_obj(account)
        return account.apiKeys
